/* slurp-download.c
 *
 * Copyright © 2017 Julian Richen <julian@richen.io>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define G_LOG_DOMAIN "slurp-download"

#include "slurp-download.h"

/**
 * SECTION:slurp-download
 * @title: SlurpDownload
 * @short_description: Asynchronous resource downloader using the libsoup library
 * @include: slurp-download.h
 * @see_also: #SoupSession, #SoupRequest, #GCallback, #GCancellable, #GFile,
 *            #GInputStream
 *
 * Asynchronous resource downloader using the libsoup library which  offers
 * addded features over using plain `soup_*` calls.
 *
 * Adds easy GSignals for when the resource download was
 * SlurpDownload::started:, SlurpDownload::finished:, and/or
 * SlurpDownload::cancelled:. You can also track the progress using the
 * SlurpDownload::progress: signal.
 **/

typedef struct
{
    SlurpResource *resource;
    gchar buffer[16384]; // 16 * 1024
    GOutputStream *output;
} SlurpDownloadTask;

typedef struct
{
    gchar *default_directory;
    gboolean default_overwrite;
    gchar *user_agent;

    SoupSession *session;
} SlurpDownloadPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (SlurpDownload, slurp_download, G_TYPE_OBJECT)

enum {
    PROP_0,
    PROP_SESSION,
    PROP_USER_AGENT,
    PROP_DEFAULT_DIRECTORY,
    PROP_DEFAULT_OVERWRITE,
    N_PROPS
};

static GParamSpec *properties[N_PROPS];

enum {
    STARTED,
    FINISHED,
    CANCELLED,
    PROGRESS,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];

gboolean
slurp_download_resource_is_valid (SlurpDownload  *self,
                                  GCancellable   *cancellable,
                                  GError        **error)
{
    g_return_val_if_fail (SLURP_IS_DOWNLOAD (self), FALSE);
    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);

    return TRUE;
}

static void
download_task_free (SlurpDownloadTask *data)
{

}

static void
download_uri_async (SlurpDownload       *self,
                    const gchar         *uri,
                    const gchar         *destination,
                    gboolean             overwrite,
                    GCancellable        *cancellable,
                    GAsyncReadyCallback  c_handler,
                    gpointer             user_data)
{
    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));

    SlurpDownloadTask *task_data = NULL;
    SlurpResource *resource = NULL;
    GTask *task = NULL;
    GError *error = NULL;
/*
    resource = g_object_new (SLURP_TYPE_RESOURCE,
                             "origin", g_strdup (uri),
                             "destination", g_strdup (destination),
                             "overwrite", overwrite,
                             NULL);

    if (!slurp_download_resource_is_valid (self, cancellable, &error))
    {
        g_task_report_error (self,
                             c_handler,
                             user_data,
                             download_uri_async,
                             error);

        g_object_unref (resource);
        return;
    }
*/

    //resource = slurp_resource_new ();

    task_data = g_slice_new0 (SlurpDownloadTask);
    //task_data->resource = resource;

    task = g_task_new (self, cancellable, c_handler, user_data);
    g_task_set_source_tag (task, download_uri_async);
    g_task_set_task_data (task, task_data, (GDestroyNotify) download_task_free);
}

gboolean
download_uri_finish (SlurpDownload  *self,
                     GAsyncResult   *result,
                     GError        **error)
{
    g_return_val_if_fail (SLURP_IS_DOWNLOAD (self), FALSE);
    g_return_val_if_fail (G_IS_TASK (result), FALSE);
    g_return_val_if_fail (g_task_is_valid (result, self), FALSE);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (result), FALSE);
    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

    GTask *task = G_TASK (result);

    g_return_val_if_fail (g_task_get_source_tag (task) == download_uri_async, FALSE);

    return g_task_propagate_boolean (task, error);
}

static void
slurp_download_finalize (GObject *object)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    g_clear_pointer (&priv->default_directory, g_free);
    g_clear_pointer (&priv->user_agent, g_free);

    G_OBJECT_CLASS (slurp_download_parent_class)->finalize (object);
}

static void
slurp_download_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SESSION:
            g_value_set_object (value, priv->session);
            break;

        case PROP_USER_AGENT:
            g_value_set_string (value, priv->user_agent);
            break;

        case PROP_DEFAULT_DIRECTORY:
            g_value_set_string (value, priv->default_directory);
            break;

        case PROP_DEFAULT_OVERWRITE:
            g_value_set_boolean (value, priv->default_overwrite);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_download_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SESSION:
            priv->session = g_value_dup_object (value);
            break;

        case PROP_USER_AGENT:
            priv->user_agent = g_value_dup_string (value);
            break;

        case PROP_DEFAULT_DIRECTORY:
            priv->default_directory = g_value_dup_string (value);
            break;

        case PROP_DEFAULT_OVERWRITE:
            priv->default_overwrite = g_value_get_boolean (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_download_class_init (SlurpDownloadClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = slurp_download_finalize;
    object_class->get_property = slurp_download_get_property;
    object_class->set_property = slurp_download_set_property;

	/**
	 * SlurpDownload:session:
	 *
	 * A SoupSession which Slurp can use to download resources.
     *
     * Note: Make sure to ref your #SoupSession before passing it.
     *
     * g_object_new():
     * |[
     * g_object_new (SLURP_TYPE_DOWNLOAD,
     *               "session", g_object_ref (session),
     *               NULL);
     * ]|
     *
     * slurp_download_new_with_session():
     * |[
     * SlurpDownload *download;
     * download = slurp_download_new_with_session (g_object_ref (session));
     * ]|
	 */
    properties[PROP_SESSION] =
        g_param_spec_object ("session",
                             "A SoupSession",
                             "A SoupSession which Slurp can use to download resources.",
                             SOUP_TYPE_SESSION,
                             G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT_ONLY);

	/**
	 * SlurpDownload:user-agent:
	 *
	 * The user agent to pass when establishing a connection to the resource.
     *
     * Default: libslurp
	 */
    // TODO: Add new user-agent if session is passed
    properties[PROP_USER_AGENT] =
        g_param_spec_string ("user-agent",
                              "User Agent",
                              "The user agent to pass when establishing a connection to the resource.",
                              "slurp-download",
                              G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT_ONLY);

	/**
	 * SlurpDownload:directory:
	 *
	 * The default directory to save resources into.
     *
     * Default: $XDG_USER_DIR/DOWNLOAD (ex. $HOME/Downloads)
	 */
    properties[PROP_DEFAULT_DIRECTORY] =
        g_param_spec_string ("directory",
                             "Default directory",
                             "The default directory to save resources into.",
                             g_get_user_special_dir (G_USER_DIRECTORY_DOWNLOAD),
                             G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT_ONLY);

	/**
	 * SlurpDownload:overwrite:
	 *
	 * Should files be overwritten by default.
     *
     * Recommended: %FALSE
	 */
    properties[PROP_DEFAULT_OVERWRITE] =
        g_param_spec_boolean ("overwrite",
                              "Overwrite files",
                              "Should files be overwritten by default? Default: FALSE",
                              FALSE,
                              G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT_ONLY);

    g_object_class_install_properties (object_class,
                                       N_PROPS,
                                       properties);

    signals[STARTED] =
        g_signal_new ("started",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0,
                      NULL);

    signals[FINISHED] =
        g_signal_new ("finished",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0,
                      NULL);
    signals[CANCELLED] =
        g_signal_new ("cancelled",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0,
                      NULL);

    signals[PROGRESS] =
        g_signal_new ("progress",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      3,
                      G_TYPE_UINT64,
                      G_TYPE_UINT64,
                      G_TYPE_UINT64);

}

static void
slurp_download_init (SlurpDownload *self)
{
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    if (!priv->session)
    {
        priv->session =
            soup_session_new_with_options (SOUP_SESSION_USER_AGENT, priv->user_agent,
                                           SOUP_SESSION_SSL_USE_SYSTEM_CA_FILE, TRUE,
                                           SOUP_SESSION_USE_THREAD_CONTEXT, TRUE,
                                           SOUP_SESSION_TIMEOUT, 60,
                                           SOUP_SESSION_IDLE_TIMEOUT, 60,
                                           NULL);
    }
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
SlurpDownload *
slurp_download_new (void)
{
    return g_object_new (SLURP_TYPE_DOWNLOAD, NULL);
}

SlurpDownload *
slurp_download_new_with_session (SoupSession *session)
{
    return g_object_new (SLURP_TYPE_DOWNLOAD,
                         "session", session,
                         NULL);
}

gboolean
slurp_download_uri (SlurpDownload *self,
                    const gchar   *uri)
{
    return TRUE;
}

gboolean
slurp_download_uri_full (SlurpDownload *self,
                         const gchar   *uri,
                         const gchar   *destination,
                         gboolean       overwrite)
{
    return TRUE;
}

void
slurp_download_uri_async (SlurpDownload       *self,
                          const gchar         *uri,
                          GCancellable        *cancellable,
                          GAsyncReadyCallback  c_handler,
                          gpointer             user_data)
{
    //download_uri_async (self, uri, destination, overwrite, cancellable, c_handler, user_data);
}

void
slurp_download_uri_async_full (SlurpDownload       *self,
                               const gchar         *uri,
                               const gchar         *destination,
                               gboolean             overwrite,
                               GCancellable        *cancellable,
                               GAsyncReadyCallback  c_handler,
                               gpointer             user_data)
{
    download_uri_async (self, uri, destination, overwrite, cancellable, c_handler, user_data);
}
