/* slurp-download.c
 *
 * Copyright © 2017 Julian Richen <julian@richen.io>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define G_LOG_DOMAIN "slurp-download"

#include "slurp-download.h"

/**
 * SECTION:slurp-download
 * @title: SlurpDownload
 * @short_description: Asynchronous resource download using the libsoup package
 * @include: slurp-download.h
 * @see_also: #SoupSession, #SoupRequest, #GMainLoop, #GMainContext,
 *            #GCallback, #GCancellable, #GFile, #GInputStream
 *
 * Downloads files in parallel by running asynchronous file download with the
 * help of libsoup. Each download runs it's own #GMainLoop but refs the current
 * #GMainContext, if one exists using g_main_context_ref_thread_default(). If no
 * current #GMainContext exists then a new one is created. If you plan on using
 * downloads inside of an application you should run that application first
 * either with a #GMainLoop or using #GApplication so I can hook onto the
 * main context.
 **/

struct _SlurpDownload
{
    GObject parent_instance;
};

typedef struct
{
    gchar *default_directory;
    gboolean default_overwrite;
    gchar *user_agent;

    GThreadPool *download_resource_thread_pool;
    SoupSession *session;
} SlurpDownloadPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (SlurpDownload, slurp_download, G_TYPE_OBJECT)

enum {
    PROP_0,
    PROP_DEFAULT_DIRECTORY,
    PROP_DEFAULT_OVERWRITE,
    PROP_USER_AGENT,
    N_PROPS
};

static GParamSpec *properties[N_PROPS];

enum {
    ADDED,
    REMOVED,
    FINISHED,
    CANCELLED,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];

/**
 * download_resource_create_path:
 * @self: a #SlurpDownload
 * @uri: a string to a uri resource
 * @path: a string to a valid path
 * @cancellable: a #GCancellable
 *
 * Create a file path to save the resource from @uri. The finial path is created
 * depending on the values of @uri, @path, and the
 * #SlurpDownload::default-directory.
 *
 * Returns: a %const string file path
 */
static const gchar *
download_resource_create_path (SlurpDownload *self,
                               const gchar   *uri,
                               const gchar   *path,
                               GCancellable  *cancellable)
{
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);
    gchar *file_path = NULL;
    gchar *file_name = NULL;
    gchar *file = NULL;

    if (uri == NULL)
        file_name = g_uuid_string_random ();
    else
        file_name = g_path_get_basename (uri);

    if (path == NULL)
        file_path = g_strdup (priv->default_directory);
    else
        file_path = g_strdup (path);

    if (g_str_has_suffix (file_path, "/"))
    {
        if (!g_file_test (file_path, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))
        {
            g_mkdir_with_parents (file_path, 0700);
        }
    }

    file = g_build_filename (file_path, file_name, NULL);

    g_free (file_path);
    g_free (file_name);

    return file;
}

/**
 * download_resource_can_write_path:
 * @self: a #SlurpDownload
 * @path: a string to a valid path
 * @overwrite: %TRUE or %FALSE if file can be overwritten
 * @cancellable: a #GCancellable
 *
 * Check if @path is writable.
 *
 * Returns: %TRUE if path is writable.
 */
static gboolean
download_resource_can_write_path (SlurpDownload *self,
                                  const gchar   *path,
                                  gboolean       overwrite,
                                  GCancellable  *cancellable)
{
    GFile *file = g_file_new_for_path (path);
    gboolean retval;

    retval = (!overwrite && g_file_query_exists (file, cancellable));

    g_object_unref (file);

    return retval;
}

/**
 * download_resource_from_uri_async_read_close_cb:
 * @object: a #GInputStream
 * @result: a #GAsyncResult
 * @user_data: a #SlurpDownloadResource
 *
 * Closes an open #GInputStream for @object.
 */
static void
download_resource_from_uri_async_read_close_cb (GObject      *object,
                                                GAsyncResult *result,
                                                gpointer      user_data)
{
    g_return_if_fail (G_IS_INPUT_STREAM (object));

    GInputStream *stream = G_INPUT_STREAM (object);
    SlurpDownloadResource *resource = user_data;
    GError *error = NULL;

    if (resource->p_handler)
    {
        /*resource->p_handler (resource->downloaded_bytes,
                         resource->total_bytes,
                         resource->p_user_data);*/
    }

    if (!g_input_stream_close_finish (stream, result, &error))
    {
        g_warning ("Downloader ( %s ): error closing GInputStream: %s", resource->uri, error->message);
    }

    g_debug ("Downloader ( %s ): closed GInputStream, read \"%lu\" bytes", resource->uri, resource->downloaded_bytes);

    slurp_download_resource_unref (resource);
}

/**
 * download_resource_from_uri_async_read_cb:
 * @object: a #GInputStream
 * @result: a #GAsyncResult
 * @user_data: a #SlurpDownloadResource
 *
 * Reads a #GInputStream from @object and saves the bytes into @user_resource->ouput
 * which is a #GOutputStream for @user_resource->file. When the steam ends
 * g_input_stream_close_async() is called with a #GCallback to
 * download_resource_from_uri_async_read_close_cb().
 */
static void
download_resource_from_uri_async_read_cb (GObject      *object,
                                          GAsyncResult *result,
                                          gpointer      user_data)
{
    g_return_if_fail (G_IS_INPUT_STREAM (object));

    GInputStream *stream = G_INPUT_STREAM (object);
    SlurpDownloadResource *resource = user_data;
    GError *error = NULL;
    gsize nread;

    if (resource->downloaded_bytes == 0)
    {
        g_debug ("Downloader ( %s ): starting async read...", resource->uri);
    }
    else
    {
        //g_debug ("Downloader ( %s ): downloaded bytes \"%lu\"", resource->uri, resource->downloaded_bytes);
    }

    nread = g_input_stream_read_finish (stream, result, &error);

    if (error)
    {
        g_warning ("Downloader ( %s ): stream read finished failed: %s",
                   resource->uri,
                   error->message);

        g_error_free (error);
    }

    if (g_get_monotonic_time () - resource->last_progress_time > 1 * G_USEC_PER_SEC)
    {
        if (resource->p_handler)
        {
            /*resource->p_handler (resource->downloaded_bytes,
                             resource->total_bytes,
                             resource->p_user_data);*/
        }

        resource->last_progress_time = g_get_monotonic_time ();
    }

    if (nread == -1 || nread == 0 || error)
    {
        // TODO: Add resource->p_handler (resource->downloaded_bytes)

        g_input_stream_close_async (stream,
                                    G_PRIORITY_DEFAULT,
                                    resource->cancellable,
                                    download_resource_from_uri_async_read_close_cb,
                                    resource);

        return;
    }

    if (resource->output)
    {
        gboolean written;
        gsize n_written;

        written = g_output_stream_write_all (G_OUTPUT_STREAM (resource->output),
                                             resource->buffer,
                                             nread,
                                             &n_written,
                                             NULL,
                                             &error);

        if (error)
        {
            g_warning ("Downloader ( %s ): stream write failed: %s",
                       resource->uri,
                       error->message);

            g_error_free (error);
        }

        resource->downloaded_bytes += n_written;

        if (!written)
        {
            g_input_stream_close_async (stream,
                                        G_PRIORITY_DEFAULT,
                                        resource->cancellable,
                                        download_resource_from_uri_async_read_close_cb,
                                        resource);
        }
    }

    g_input_stream_read_async (stream,
                               resource->buffer,
                               sizeof (resource->buffer),
                               G_PRIORITY_DEFAULT,
                               resource->cancellable,
                               download_resource_from_uri_async_read_cb,
                               resource);
}

/**
 * download_resource_from_uri_async_cb:
 * @object: a #SoupRequest
 * @result: a #GAsyncResult
 * @user_data: a #SlurpDownloadResource
 *
 * After a successful #SoupRequest async callback a #GInputStream is returned.
 * If the #GInputStream is error free then a #GOutputStream is created for the
 * #GFile at user_resource->file. Assuming a #GOutputStream is created without
 * errors then g_input_stream_read_async() is called with a #GCallback to
 * download_resource_from_uri_async_read_cb() which reads the #GInputStream of
 * the #SoupRequest.
 */
static void
download_resource_from_uri_async_cb (GObject      *object,
                                     GAsyncResult *result,
                                     gpointer      user_data)
{
    g_return_if_fail (SOUP_IS_REQUEST (object));

    SoupRequest *request = SOUP_REQUEST (object);
    GInputStream *stream = NULL;
    SlurpDownloadResource *resource = user_data;
    GError *error = NULL;

    g_debug ("Downloader ( %s ): starting async fetch...", resource->uri);

    stream = soup_request_send_finish (request, result, &error);

    resource->total_bytes = soup_request_get_content_length (request);

    if (error)
    {
        g_warning ("Downloader ( %s ): failed to start resource: %s",
                   resource->uri,
                   error->message);

        g_error_free (error);
        slurp_download_resource_unref (resource);

        return;
    }

    resource->output = G_OUTPUT_STREAM (g_file_replace (resource->file,
                                                        NULL,
                                                        FALSE,
                                                        G_FILE_CREATE_NONE,
                                                        resource->cancellable,
                                                        &error));

    if (error)
    {
        g_warning ("Downloader ( %s ): failed to create output stream and file \"%s\": %s",
                   resource->uri,
                   resource->path,
                   error->message);

        g_error_free (error);
        slurp_download_resource_unref (resource);

        return;
    }

    g_input_stream_read_async (stream,
                               resource->buffer,
                               sizeof (resource->buffer),
                               G_PRIORITY_DEFAULT,
                               resource->cancellable,
                               download_resource_from_uri_async_read_cb,
                               resource);
}

/**
 * download_resource_from_uri_async_full:
 * @self: A #SlurpDownload
 * @uri: a uri to a resource to download
 * @cancellable: a #GCancellable, used to stop the asynchronous resource fetch
 *               and resource save
 * @overwrite: %TRUE to overwrite @path
 * @path: a path to save the @uri resource to
 * @progress: the #GCallback for the progress callback.
 *
 * Requests a resource from @uri saving it into @path, overriding the contents
 * of @path if @overwrite is set to %TRUE. If @path is %NULL then the @uri will
 * be downloaded into %G_USER_DIRECTORY_DOWNLOAD (a.k.a ~/Downloads). If path
 * ends in a slash (/) then the path is treated as a directory. In both cases
 * the file will be named the basename of @uri. Not supplying @path is
 * discouraged.
 *
 * In addition a #GFile is created for @path and set into a new
 * #SlurpDownloadResource struct as (struct name)->file. Checks are also made if
 * the @path file exists and if it should overwrite based on @overwrite.
 *
 * When initial data is set soup_request_send_async() is called with a
 * #GCallback to download_resource_from_uri_async_cb() which starts the
 * #SoupRequest.
 *
 * Note: download_resource_from_uri_async_full() runs it's own #GMainLoop
 * but refs the current #GMainContext, if on exists using
 * g_main_context_ref_thread_default(). If no current #GMainContext exists then
 * a new one is created. If you plan on using downloads inside of an
 * application you should run that application first either with a #GMainLoop
 * or using #GApplication so I can hook onto the main context.
 */
static void
download_resource_from_uri_async_full (SlurpDownload *self,
                                       const gchar   *uri,
                                       GCancellable  *cancellable,
                                       gboolean       overwrite,
                                       const gchar   *path,
                                       GCallback      p_handler,
                                       gpointer       p_user_data)
{
    g_return_if_fail (uri);
    g_return_if_fail (!cancellable || G_IS_CANCELLABLE (cancellable));

    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);
    GMainLoop *loop = NULL;
    GMainContext *context = NULL;
    SlurpDownloadResource *resource;
    SoupRequest *request;
    const gchar *internal_path = NULL;
    GError *error = NULL;

    context = g_main_context_ref_thread_default ();
    loop = g_main_loop_new (context, TRUE);

    g_debug ("Downloader ( %s ): starting...", uri);

    internal_path = download_resource_create_path (self, uri, path, cancellable);

    g_debug ("Downloader ( %s ): saving to \"%s\"", uri, internal_path);

    if (download_resource_can_write_path (self, internal_path, overwrite, cancellable))
    {
        g_debug ("Downloader ( %s ): overwite = FALSE and file exists ( %s ), download cancelled", uri, internal_path);
        g_cancellable_cancel (cancellable);
        g_clear_pointer (&cancellable, g_object_unref);
        return;
    }

    request = soup_session_request (priv->session, uri, &error);

    if (error)
    {
        g_warning ("Downloader ( %s ): failed to prase uri", uri);
        g_cancellable_cancel (cancellable);
        g_clear_pointer (&cancellable, g_object_unref);
        g_clear_error (&error);
        g_error_free (error);
        return;
    }

    g_debug ("Downloader ( %s ): soup session and request started", uri);

    resource = slurp_download_resource_new ();
    resource->loop = loop;

    resource->uri = uri;
    resource->path = internal_path;
    resource->overwrite = overwrite;

    resource->cancellable = cancellable;
    resource->p_handler = p_handler;
    resource->p_user_data = p_user_data;

    resource->file = g_file_new_for_path (resource->path);
    resource->total_bytes = 0;
    resource->downloaded_bytes = 0;
    resource->last_progress_time = g_get_monotonic_time ();

    soup_request_send_async (SOUP_REQUEST(request),
                             cancellable,
                             download_resource_from_uri_async_cb,
                             resource);

    g_main_loop_run (loop);
    g_main_loop_unref (loop);
    g_main_context_unref (context);
}















typedef struct _SlurpDownloadTaskData {
    gchar *uuid;
    const gchar *uri;
    const gchar *path;
    gboolean overwrite;

    GFile *file;
    gchar buffer[16384]; // 16 * 1024
    GOutputStream *output;

    guint64 total_bytes;
    guint64 downloaded_bytes;
    guint64 last_progress_time;
} SlurpDownloadTaskData;

static void
slurp_download_task_data_free (SlurpDownloadTaskData *data)
{
    g_free (data->uuid);
    g_clear_pointer (&data->file, g_object_unref);
    g_free (data->buffer);
    g_clear_pointer (&data->output, g_object_unref);

    g_slice_free (SlurpDownloadTaskData, data);
}

static void
_download_resource_from_uri_async ()
{
}

gboolean
slurp_download_resource_from_uri (SlurpDownload  *self,
                                  const gchar    *uri,
                                  const gchar    *path,
                                  gboolean        overwrite,
                                  GCancellable   *cancellable,
                                  GError        **error)
{
    g_print ("Downloading...\n");

    // Check URI
    // Check path
    // Check write
    // Start download

    soup_request_send_async (SOUP_REQUEST(request),
                             cancellable,
                             download_resource_from_uri_async_cb,
                             resource);

    return TRUE;
}

static void
slurp_download_resource_in_thread_func (gpointer data,      /* task_data from g_task_new() */
                                        gpointer user_data) /* user_data from g_thread_pool_new() */
{
    g_return_if_fail (G_IS_TASK (data) && data != NULL);
    g_return_if_fail (SLURP_IS_DOWNLOAD (user_data));

    GTask *task = G_TASK (data);
    SlurpDownload *self = SLURP_DOWNLOAD (user_data);
    SlurpDownloadTaskData *task_data = g_task_get_task_data (task);
    GCancellable *cancellable = g_task_get_cancellable (task);
    GError *error = NULL;
    gboolean retval;

    retval = slurp_download_resource_from_uri (self,
                                               task_data->uri,
                                               task_data->path,
                                               task_data->overwrite,
                                               cancellable,
                                               &error);

    if (!retval)
    {
        /*
         * If the user cancelled the download then clean-up the file
         */
        if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
            g_file_delete (task_data->file, NULL, NULL);

        g_task_return_error (task, error);
        return;
    }

    g_task_return_boolean (task, TRUE);
}

void
slurp_download_resource_from_uri_async (SlurpDownload       *self,
                                        const gchar         *uri,
                                        const gchar         *path,
                                        gboolean             overwrite,
                                        GCancellable        *cancellable,
                                        GAsyncReadyCallback  callback,
                                        gpointer             user_data)
{
    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));

    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);
    SlurpDownloadTaskData *task_data = NULL;
    GTask *task = NULL;
    GError *error = NULL;
    gboolean retval = FALSE;

    task_data = g_slice_new0 (SlurpDownloadTaskData);
    task_data->uuid = g_uuid_string_random ();
    task_data->uri = g_strdup (uri);
    task_data->overwrite = overwrite;
    task_data->path = g_strdup (path);
    task_data->file = NULL;

    task_data->total_bytes = 0;
    task_data->downloaded_bytes = 0;
    task_data->last_progress_time = g_get_monotonic_time ();

    g_debug ("Resource added to download queue with UUID \"%s\"\n from URI \"%s\"",
             task_data->uuid,
             task_data->uri);

    task = g_task_new (self, cancellable, callback, user_data);
    g_task_set_source_tag (task, slurp_download_resource_from_uri_async);
    g_task_set_task_data (task, task_data, (GDestroyNotify) slurp_download_task_data_free);

    retval = g_thread_pool_push (priv->download_resource_thread_pool,
                                 g_object_ref (task),
                                 &error);

    if (!retval)
    {
        g_debug ("Unable to add resource with UUID \"%s\" to download queue", task_data->uuid);
        g_object_unref (task);
        return;
    }
}

gboolean
slurp_download_resource_from_uri_finish (SlurpDownload  *self,
                                         GAsyncResult   *result,
                                         GError        **error)
{
    g_return_val_if_fail (SLURP_IS_DOWNLOAD (self), FALSE);
    g_return_val_if_fail (G_IS_TASK (result), FALSE);
    g_return_val_if_fail (g_task_is_valid (result, self), FALSE);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (result), FALSE);
    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

    GTask *task = G_TASK (result);

    g_return_val_if_fail (g_task_get_source_tag (task) == slurp_download_resource_from_uri_async, FALSE);

    return g_task_propagate_boolean (task, error);
}

static void
slurp_download_resource_from_uri_cb (GObject      *object,
                                     GAsyncResult *result,
                                     gpointer      user_data)
{
    g_return_if_fail (SLURP_IS_DOWNLOAD (object));
    g_return_if_fail (G_IS_ASYNC_RESULT (result));

    SlurpDownload *self = SLURP_DOWNLOAD (object);
    gboolean retval;
    GError *error = NULL;

    retval = slurp_download_resource_from_uri_finish (self, result, &error);

    if (error != NULL)
    {
        if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
            goto out;

        // TODO: make better
        g_warning ("Unable to download resource: %s", error->message);
    }

    if (!retval)
    {
        //TODO: Trigger failed download signal
        goto out;
    }

    //TODO: Trigger success download signal

out:
    return;
}

void
slurp_download_add_uri_full (SlurpDownload *self,
                             const gchar   *uri,
                             const gchar   *path,
                             gboolean       overwrite,
                             GCancellable  *cancellable,
                             GCallback      p_handler,
                             gpointer       p_user_data)
{
    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));

    slurp_download_resource_from_uri_async (self,
                                            uri,
                                            path,
                                            overwrite,
                                            cancellable,
                                            slurp_download_resource_from_uri_cb,
                                            NULL);
}

void
slurp_download_add_uri (SlurpDownload *self,
                        const gchar   *uri,
                        GCancellable  *cancellable)
{
    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));

    slurp_download_add_uri_full (self,
                                 uri,
                                 NULL,
                                 FALSE,
                                 cancellable,
                                 NULL,
                                 NULL);
}

static void
slurp_download_finalize (GObject *object)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    g_clear_pointer (&priv->default_directory, g_free);
    g_clear_pointer (&priv->user_agent, g_free);

    G_OBJECT_CLASS (slurp_download_parent_class)->finalize (object);
}

static void
slurp_download_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_DEFAULT_DIRECTORY:
            g_value_set_string (value, priv->default_directory);
            break;

        case PROP_DEFAULT_OVERWRITE:
            g_value_set_boolean (value, priv->default_overwrite);
            break;

        case PROP_USER_AGENT:
            g_value_set_string (value, priv->user_agent);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_download_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_DEFAULT_DIRECTORY:
            priv->default_directory = g_value_dup_string (value);
            break;

        case PROP_DEFAULT_OVERWRITE:
            priv->default_overwrite = g_value_get_boolean (value);
            break;

        case PROP_USER_AGENT:
            priv->user_agent = g_value_dup_string (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_download_class_init (SlurpDownloadClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = slurp_download_finalize;
    object_class->get_property = slurp_download_get_property;
    object_class->set_property = slurp_download_set_property;

	/**
	 * SlurpDownload:directory:
	 *
	 * The default directory to save resources into.
     *
     * Default: $XDG_USER_DIR/DOWNLOAD (ex. $HOME/Downloads)
	 */
    properties[PROP_DEFAULT_DIRECTORY] =
        g_param_spec_string ("directory",
                             "Default directory",
                             "The default directory to save resources into.",
                             g_get_user_special_dir (G_USER_DIRECTORY_DOWNLOAD),
                             G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT_ONLY);

	/**
	 * SlurpDownload:overwrite:
	 *
	 * Should files be overwritten by default.
     *
     * Recommended: %FALSE
	 */
    properties[PROP_DEFAULT_OVERWRITE] =
        g_param_spec_boolean ("overwrite",
                              "Overwrite files",
                              "Should files be overwritten by default? Default: FALSE",
                              FALSE,
                              G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT_ONLY);

	/**
	 * SlurpDownload:user-agent:
	 *
	 * The user agent to pass when establishing a connection to the resource.
     *
     * Default: libslurp
	 */
    properties[PROP_USER_AGENT] =
        g_param_spec_string ("user-agent",
                              "User Agent",
                              "The user agent to pass when establishing a connection to the resource.",
                              "libslurp",
                              G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT_ONLY);

    g_object_class_install_properties (object_class,
                                       N_PROPS,
                                       properties);

    signals[ADDED] =
        g_signal_new ("added",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_STRING);

    signals[REMOVED] =
        g_signal_new ("removed",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_STRING);

    signals[FINISHED] =
        g_signal_new ("finished",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_STRING);

    signals[CANCELLED] =
        g_signal_new ("cancelled",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_STRING);
}

static void
slurp_download_init (SlurpDownload *self)
{
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    priv->download_resource_thread_pool =
        g_thread_pool_new (slurp_download_resource_in_thread_func,
                           self,
                           1,
                           FALSE,
                           NULL);

    priv->session =
        soup_session_new_with_options (SOUP_SESSION_USER_AGENT, priv->user_agent,
                                       SOUP_SESSION_SSL_USE_SYSTEM_CA_FILE, TRUE,
                                       SOUP_SESSION_USE_THREAD_CONTEXT, TRUE,
                                       SOUP_SESSION_TIMEOUT, 60,
                                       SOUP_SESSION_IDLE_TIMEOUT, 60,
                                       NULL);
}

SlurpDownload *
slurp_download_new (void)
{
    return g_object_new (SLURP_TYPE_DOWNLOAD, NULL);
}
